@@ -1,6 +1,6 @@
 /*
 
-Copyright (C) 2013 Olaf Till <i7tiol@t-online.de>
+Copyright (C) 2013-2016 Olaf Till <i7tiol@t-online.de>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -24,11 +24,15 @@ along with this program; If not, see <http://www.gnu.org/licenses/>.
 #include <stdio.h>
 
 #include "command.h"
+#include "error-helpers.h"
 #include <libpq/libpq-fs.h>
 
 // PKG_ADD: autoload ("pq_lo_import", "pq_interface.oct");
 // PKG_ADD: autoload ("pq_lo_export", "pq_interface.oct");
 // PKG_ADD: autoload ("pq_lo_unlink", "pq_interface.oct");
+// PKG_DEL: autoload ("pq_lo_import", "pq_interface.oct", "remove");
+// PKG_DEL: autoload ("pq_lo_export", "pq_interface.oct", "remove");
+// PKG_DEL: autoload ("pq_lo_unlink", "pq_interface.oct", "remove");
 
 #define OCT_PQ_BUFSIZE 1024
 
@@ -37,7 +41,7 @@ class pipe_to_lo
 {
 public:
 
-  pipe_to_lo (octave_pq_connection &, const char *, bool, std::string &);
+  pipe_to_lo (octave_pq_connection_rep &, const char *, bool, std::string &);
 
   ~pipe_to_lo (void);
 
@@ -49,7 +53,7 @@ public:
 
 private:
 
-  octave_pq_connection &oct_pq_conn;
+  octave_pq_connection_rep &oct_pq_conn;
 
   PGconn *conn;
 
@@ -64,7 +68,7 @@ private:
   bool commit;
 };
 
-pipe_to_lo::pipe_to_lo (octave_pq_connection &a_oct_pq_conn,
+pipe_to_lo::pipe_to_lo (octave_pq_connection_rep &a_oct_pq_conn,
                         const char *cmd, bool acommit, std::string &amsg)
   : msg (amsg), oct_pq_conn (a_oct_pq_conn),
     conn (a_oct_pq_conn.octave_pq_get_conn ()), oid (0), fp (NULL),
@@ -125,7 +129,7 @@ pipe_to_lo::pipe_to_lo (octave_pq_connection &a_oct_pq_conn,
   if (nb) return;
 
   if (pclose (fp) == -1)
-    error ("error closing pipe");
+    c_verror ("error closing pipe");
 
   fp = NULL;
 
@@ -142,7 +146,7 @@ pipe_to_lo::~pipe_to_lo (void)
   if (lod != -1)
     {
       if (lo_close (conn, lod))
-        error ("%s", PQerrorMessage (conn));
+        c_verror ("%s", PQerrorMessage (conn));
 
       lod = -1;
     }
@@ -150,7 +154,7 @@ pipe_to_lo::~pipe_to_lo (void)
   if (oid && ! oid_valid)
     {
       if (lo_unlink (conn, oid) == -1)
-        error ("error unlinking new large object with oid %i", oid);
+        c_verror ("error unlinking new large object with oid %i", oid);
     }
   else
     oid = 0;
@@ -158,7 +162,7 @@ pipe_to_lo::~pipe_to_lo (void)
   if (fp)
     {
       if (pclose (fp) == -1)
-        error ("error closing pipe");
+        c_verror ("error closing pipe");
 
       fp = NULL;
     }
@@ -176,7 +180,7 @@ pipe_to_lo::~pipe_to_lo (void)
         c.process_single_result ();
 
       if (! c.good ())
-        error ("pq_lo_import: could not commit");
+        c_verror ("%s: could not commit", caller.c_str ());
     }
 }
 
@@ -185,7 +189,7 @@ class lo_to_pipe
 {
 public:
 
-  lo_to_pipe (octave_pq_connection &, Oid, const char *, bool, std::string &);
+  lo_to_pipe (octave_pq_connection_rep &, Oid, const char *, bool, std::string &);
 
   ~lo_to_pipe (void);
 
@@ -195,7 +199,7 @@ public:
 
 private:
 
-  octave_pq_connection &oct_pq_conn;
+  octave_pq_connection_rep &oct_pq_conn;
 
   PGconn *conn;
 
@@ -210,7 +214,7 @@ private:
   bool commit;
 };
 
-lo_to_pipe::lo_to_pipe (octave_pq_connection &a_oct_pq_conn, Oid aoid,
+lo_to_pipe::lo_to_pipe (octave_pq_connection_rep &a_oct_pq_conn, Oid aoid,
                         const char *cmd, bool acommit, std::string &amsg) :
   msg (amsg), oct_pq_conn (a_oct_pq_conn),
   conn (a_oct_pq_conn.octave_pq_get_conn ()), oid (aoid), fp (NULL),
@@ -266,7 +270,7 @@ lo_to_pipe::lo_to_pipe (octave_pq_connection &a_oct_pq_conn, Oid aoid,
   if (pnb) return;
 
   if (pclose (fp) == -1)
-    error ("error closing pipe");
+    c_verror ("error closing pipe");
 
   fp = NULL;
 
@@ -283,7 +287,7 @@ lo_to_pipe::~lo_to_pipe (void)
   if (lod != -1)
     {
       if (lo_close (conn, lod))
-        error ("%s", PQerrorMessage (conn));
+        c_verror ("%s", PQerrorMessage (conn));
 
       lod = -1;
     }
@@ -291,7 +295,7 @@ lo_to_pipe::~lo_to_pipe (void)
   if (fp)
     {
       if (pclose (fp) == -1)
-        error ("error closing pipe");
+        c_verror ("error closing pipe");
 
       fp = NULL;
     }
@@ -309,7 +313,7 @@ lo_to_pipe::~lo_to_pipe (void)
         c.process_single_result ();
 
       if (! c.good ())
-        error ("pq_lo_export: could not commit");
+        c_verror ("%s: could not commit", caller.c_str ());
     }
 }
 
@@ -331,15 +335,10 @@ Imports the file in @var{path} on the client side as a large object into the dat
       return retval;
     }
 
-  std::string path (args(1).string_value ());
-
-  if (error_state)
-    {
-      error ("%s: second argument can not be converted to a string",
-             fname.c_str ());
-
-      return retval;
-    }
+  std::string path;
+  CHECK_ERROR (path = args(1).string_value (), retval,
+               "%s: second argument can not be converted to a string",
+               fname.c_str ());
 
   bool from_pipe = false;
   unsigned int l = path.size ();
@@ -362,11 +361,12 @@ Imports the file in @var{path} on the client side as a large object into the dat
       from_pipe = true;
     }
 
-  octave_base_value& rep = const_cast<octave_base_value&> (args(0).get_rep ());
+  const octave_base_value& rep = (args(0).get_rep ());
 
-  octave_pq_connection &oct_pq_conn = dynamic_cast<octave_pq_connection&> (rep);
+  const octave_pq_connection &oct_pq_conn =
+    dynamic_cast<const octave_pq_connection&> (rep);
 
-  PGconn *conn = oct_pq_conn.octave_pq_get_conn ();
+  PGconn *conn = oct_pq_conn.get_rep ()->octave_pq_get_conn ();
 
   if (! conn)
     {
@@ -385,24 +385,22 @@ Imports the file in @var{path} on the client side as a large object into the dat
     case PQTRANS_INERROR:
       error ("%s: can't manipulate large objects within a failed transaction block",
              fname.c_str ());
-      break;
+      return retval;
     case PQTRANS_UNKNOWN:
       error ("%s: connection is bad", fname.c_str ());
-      break;
+      return retval;
     default: // includes PQTRANS_ACTIVE
       error ("%s: unexpected connection state", fname.c_str ());
+      return retval;
     }
 
-  if (error_state)
-    return retval;
-
   if (make_tblock)
     {
       std::string cmd ("begin;");
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -421,7 +419,7 @@ Imports the file in @var{path} on the client side as a large object into the dat
 
   if (from_pipe)
     {
-      pipe_to_lo tp (oct_pq_conn, path.c_str (), make_tblock, msg);
+      pipe_to_lo tp (*(oct_pq_conn.get_rep ()), path.c_str (), make_tblock, msg);
 
       make_tblock = false; // commit handled by destructor of pipe_to_lo
 
@@ -445,7 +443,7 @@ Imports the file in @var{path} on the client side as a large object into the dat
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -455,13 +453,17 @@ Imports the file in @var{path} on the client side as a large object into the dat
     }
 
   if (import_error)
-    error ("%s: large object import failed: %s", fname.c_str (), msg.c_str ());
+    c_verror ("%s: large object import failed: %s",
+              fname.c_str (), msg.c_str ());
 
   if (commit_error)
-    error ("%s: could not commit transaction", fname.c_str ());
+    c_verror ("%s: could not commit transaction", fname.c_str ());
 
-  if (error_state)
-    return retval;
+  if (import_error || commit_error)
+    {
+      error ("%s failed", fname.c_str ());
+      return retval;
+    }
 
   retval = octave_value (octave_uint32 (oid));
 
@@ -487,15 +489,10 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
       return retval;
     }
 
-  std::string path (args(2).string_value ());
-
-  if (error_state)
-    {
-      error ("%s: third argument can not be converted to a string",
-             fname.c_str ());
-
-      return retval;
-    }
+  std::string path;
+  CHECK_ERROR (path = args(2).string_value (), retval,
+               "%s: third argument can not be converted to a string",
+               fname.c_str ());
 
   bool to_pipe = false;
   if (! path.empty () && path[0] == '|')
@@ -513,21 +510,17 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
       to_pipe = true;
     }
 
-  Oid oid = args(1).uint_value ();
+  Oid oid;
+  CHECK_ERROR (oid = args(1).uint_value (), retval,
+               "%s: second argument can not be converted to an oid",
+               fname.c_str ());
 
-  if (error_state)
-    {
-      error ("%s: second argument can not be converted to an oid",
-             fname.c_str ());
+  const octave_base_value& rep = (args(0).get_rep ());
 
-      return retval;
-    }
+  const octave_pq_connection &oct_pq_conn =
+    dynamic_cast<const octave_pq_connection&> (rep);
 
-  octave_base_value& rep = const_cast<octave_base_value&> (args(0).get_rep ());
-
-  octave_pq_connection &oct_pq_conn = dynamic_cast<octave_pq_connection&> (rep);
-
-  PGconn *conn = oct_pq_conn.octave_pq_get_conn ();
+  PGconn *conn = oct_pq_conn.get_rep ()->octave_pq_get_conn ();
 
   if (! conn)
     {
@@ -546,24 +539,22 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
     case PQTRANS_INERROR:
       error ("%s: can't manipulate large objects within a failed transaction block",
              fname.c_str ());
-      break;
+      return retval;
     case PQTRANS_UNKNOWN:
       error ("%s: connection is bad", fname.c_str ());
-      break;
+      return retval;
     default: // includes PQTRANS_ACTIVE
       error ("%s: unexpected connection state", fname.c_str ());
+      return retval;
     }
 
-  if (error_state)
-    return retval;
-
   if (make_tblock)
     {
       std::string cmd ("begin;");
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -580,7 +571,7 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
 
   if (to_pipe)
     {
-      lo_to_pipe tp (oct_pq_conn, oid, path.c_str (), make_tblock, msg);
+      lo_to_pipe tp (*(oct_pq_conn.get_rep ()), oid, path.c_str (), make_tblock, msg);
 
       make_tblock = false; // commit handled by destructor of lo_to_pipe
 
@@ -602,7 +593,7 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -612,10 +603,14 @@ Exports the large object of Oid @var{oid} in the database associated with @var{c
     }
 
   if (export_error)
-    error ("%s: large object export failed: %s", fname.c_str (), msg.c_str ());
+    c_verror ("%s: large object export failed: %s",
+              fname.c_str (), msg.c_str ());
 
   if (commit_error)
-    error ("%s: could not commit transaction", fname.c_str ());
+    c_verror ("%s: could not commit transaction", fname.c_str ());
+
+  if (export_error || commit_error)
+    error ("%s failed", fname.c_str ());
 
   return retval;
 }
@@ -639,21 +634,17 @@ Removes the large object of Oid @var{oid} from the database associated with @var
       return retval;
     }
 
-  Oid oid = args(1).uint_value ();
+  Oid oid;
+  CHECK_ERROR (oid = args(1).uint_value (), retval,
+               "%s: second argument can not be converted to an oid",
+               fname.c_str ());
 
-  if (error_state)
-    {
-      error ("%s: second argument can not be converted to an oid",
-             fname.c_str ());
+  const octave_base_value& rep = (args(0).get_rep ());
 
-      return retval;
-    }
+  const octave_pq_connection &oct_pq_conn =
+    dynamic_cast<const octave_pq_connection&> (rep);
 
-  octave_base_value& rep = const_cast<octave_base_value&> (args(0).get_rep ());
-
-  octave_pq_connection &oct_pq_conn = dynamic_cast<octave_pq_connection&> (rep);
-
-  PGconn *conn = oct_pq_conn.octave_pq_get_conn ();
+  PGconn *conn = oct_pq_conn.get_rep ()->octave_pq_get_conn ();
 
   if (! conn)
     {
@@ -672,24 +663,22 @@ Removes the large object of Oid @var{oid} from the database associated with @var
     case PQTRANS_INERROR:
       error ("%s: can't manipulate large objects within a failed transaction block",
              fname.c_str ());
-      break;
+      return retval;
     case PQTRANS_UNKNOWN:
       error ("%s: connection is bad", fname.c_str ());
-      break;
+      return retval;
     default: // includes PQTRANS_ACTIVE
       error ("%s: unexpected connection state", fname.c_str ());
+      return retval;
     }
 
-  if (error_state)
-    return retval;
-
   if (make_tblock)
     {
       std::string cmd ("begin;");
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -718,7 +707,7 @@ Removes the large object of Oid @var{oid} from the database associated with @var
       Cell params;
       Cell ptypes (1, 0);
       Cell rtypes;
-      command c (oct_pq_conn, cmd, params, ptypes, rtypes, fname);
+      command c (*(oct_pq_conn.get_rep ()), cmd, params, ptypes, rtypes, fname);
 
       if (c.good ())
         c.process_single_result ();
@@ -728,10 +717,14 @@ Removes the large object of Oid @var{oid} from the database associated with @var
     }
 
   if (unlink_error)
-    error ("%s: large object unlink failed: %s", fname.c_str (), msg.c_str ());
+    c_verror ("%s: large object unlink failed: %s",
+              fname.c_str (), msg.c_str ());
 
   if (commit_error)
-    error ("%s: could not commit transaction", fname.c_str ());
+    c_verror ("%s: could not commit transaction", fname.c_str ());
+
+  if (unlink_error || commit_error)
+    error ("%s failed", fname.c_str ());      
 
   return retval;
 }
